import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/ast/visitor.dart';
import 'package:analyzer_plugin/protocol/protocol_common.dart' as plugin;
import 'package:analyzer_plugin/protocol/protocol_common.dart';
import 'package:pt_linter/logger/log.dart';
import 'package:analyzer_plugin/protocol/protocol_generated.dart';
import 'package:analyzer/src/dart/ast/constant_evaluator.dart';

class MirrorChecker {
  final CompilationUnit _compilationUnit;
  String unitPath;

  MirrorChecker(this._compilationUnit) {
    unitPath = this._compilationUnit.declaredElement.source.fullName;
    // mirrorLog.info("checker $unitPath");
  }

  Map<MirrorCheckerIssue, PrioritizedSourceChange> enumToStringErrors() {
    final visitor = _MirrorVisitor(_compilationUnit, unitPath);
    _compilationUnit.accept(visitor);
    return visitor.results;
  }
}

class _MirrorVisitor extends RecursiveAstVisitor<void> {
  _MirrorVisitor(this.compilationUnit, this.unitPath);

  final CompilationUnit compilationUnit;
  final String unitPath;
  final _results = <MirrorCheckerIssue, PrioritizedSourceChange>{};
  static final _evaluator = new ConstantEvaluator();

  Map<MirrorCheckerIssue, PrioritizedSourceChange> get results => _results;

  @override
  void visitInstanceCreationExpression(InstanceCreationExpression node) {
    node.visitChildren(this);
    final constructorName = node.constructorName.toString();

    if (constructorName == 'Border.only' ||
        constructorName == 'BorderRadius.only' ||
        constructorName == 'BorderRadius.horizontal' ||
        constructorName == 'Positioned' ||
        constructorName == 'Positioned.fill' ||
        constructorName == 'EdgeInsets' ||
        constructorName == 'EdgeInsets.only' ||
        constructorName == 'EdgeInsets.fromLTRB') {
      bool shouldFix = false;
      List<String> arguments = [];

      node.argumentList.arguments.forEach((element) {
        if (element is! NamedExpression) return;

        NamedExpression express = element as NamedExpression;
        var result = express.expression.accept(_evaluator);
        if (result == ConstantEvaluator.NOT_A_CONSTANT) {
          result = express.expression.toString();
        }

        String label = express.name.toString();
        if (label == 'left:') {
          label = 'start:';
          shouldFix = true;
        }
        if (label == 'right:') {
          label = 'end:';
          shouldFix = true;
        }
        String split = ' ';

        arguments.add("$label$split$result");
      });
      if (!shouldFix) return;

      final constructorNameType = node.constructorName.type.toString();
      String fixResult = constructorName;

      fixResult = fixResult.replaceFirst(constructorNameType, "${constructorNameType}Directional");

      if ('Positioned.fill' == constructorName) fixResult = 'PositionedDirectional';
      if ('EdgeInsets.fromLTRB' == constructorName) fixResult = 'EdgeInsetsDirectional.fromSTEB';

      fixResult += '( ';
      fixResult += arguments.join(',');
      fixResult += ' )';

      final issue = MirrorCheckerIssue(
        plugin.AnalysisErrorSeverity.ERROR,
        plugin.AnalysisErrorType.LINT,
        node.offset,
        node.length,
        'RTL:Use ${constructorNameType}Directional instead',
        '$constructorName is not RTL campatiable',
      );

      final fix = PrioritizedSourceChange(
          1000000,
          SourceChange(
            'Use ${constructorNameType}Directional instead',
            edits: [
              SourceFileEdit(
                compilationUnit.declaredElement.source.fullName,
                compilationUnit.declaredElement.source.modificationStamp,
                edits: [SourceEdit(node.offset, node.length, fixResult)],
              )
            ],
          ));
      _results[issue] = fix;
    }
  }

  @override
  void visitSimpleStringLiteral(SimpleStringLiteral node) {
    node.visitChildren(this);
    if(!_shouldFixText(node)) return;

    if (node.value != null && node.value.contains(RegExp(r'([\u4e00-\u9fa5]+)')) == false) return;

    final issue = MirrorCheckerIssue(
      plugin.AnalysisErrorSeverity.ERROR,
      plugin.AnalysisErrorType.HINT,
      node.offset,
      node.length,
      'Hardcoded string is not intl compatible',
      'hardcoded_string',
    );

    final fix = PrioritizedSourceChange(
        1000000,
        SourceChange(
          'Extract to K.',
          edits: [
            SourceFileEdit(
              compilationUnit.declaredElement.source.fullName,
              compilationUnit.declaredElement.source.modificationStamp,
              edits: [],
            )
          ],
        ));
    _results[issue] = fix;
  }

  @override
  void visitStringInterpolation(StringInterpolation node) {
    node.visitChildren(this);
    if(!_shouldFixText(node)) return;
    
    if(node.toString().isNotEmpty && node.toString().contains(RegExp(r'([\u4e00-\u9fa5]+)')) == false) return;
    

    final issue = MirrorCheckerIssue(
      plugin.AnalysisErrorSeverity.ERROR,
      plugin.AnalysisErrorType.HINT,
      node.offset,
      node.length,
      'Hardcoded string is not intl compatible',
      'hardcoded_string',
    );

    final fix = PrioritizedSourceChange(
        1000000,
        SourceChange(
          'Extract to K.',
          edits: [
            SourceFileEdit(
              compilationUnit.declaredElement.source.fullName,
              compilationUnit.declaredElement.source.modificationStamp,
              edits: [],
            )
          ],
        ));
    _results[issue] = fix;
  }

  @override
  void visitPrefixedIdentifier(PrefixedIdentifier node) {
    node.visitChildren(this);

    final prefix = node.prefix.name;
    if ('Alignment' == prefix) {
      bool shouldFix = false;
      String fixResult = node.toString().replaceFirst(prefix, "${prefix}Directional");
      if (node.identifier.name.contains('Left')) {
        fixResult = fixResult.replaceFirst('Left', 'Start');
        shouldFix = true;
      }

      if (node.identifier.name.contains('Right')) {
        fixResult = fixResult.replaceFirst('Right', 'End');
        shouldFix = true;
      }

      if (!shouldFix) return;

      final issue = MirrorCheckerIssue(
        plugin.AnalysisErrorSeverity.ERROR,
        plugin.AnalysisErrorType.LINT,
        node.offset,
        node.length,
        'RTL:Use ${prefix}Directional instead',
        '$prefix is not RTL campatiable',
      );

      final fix = PrioritizedSourceChange(
          1000000,
          SourceChange(
            'Use ${prefix}Directional instead',
            edits: [
              SourceFileEdit(
                compilationUnit.declaredElement.source.fullName,
                compilationUnit.declaredElement.source.modificationStamp,
                edits: [SourceEdit(node.offset, node.length, fixResult)],
              )
            ],
          ));
      _results[issue] = fix;
    }
  }

  bool _shouldFixText(AstNode node) {
    if(node == null || node.parent == null) return false;

    if(node.parent is ArgumentList && node.parent.parent is InstanceCreationExpression){
      return true;
      //String constructorName = (node.parent.parent as InstanceCreationExpression).constructorName.toString();
      //if(constructorName == 'Text' || constructorName == 'RichText' || constructorName == 'RichText')
    }

    if(node.parent is ReturnStatement || node.parent is AssignmentExpression) return true;

    return false;
  }
}

class MirrorCheckerIssue {
  final plugin.AnalysisErrorSeverity analysisErrorSeverity;
  final plugin.AnalysisErrorType analysisErrorType;
  final int offset;
  final int length;
  final String message;
  final String code;

  MirrorCheckerIssue(
    this.analysisErrorSeverity,
    this.analysisErrorType,
    this.offset,
    this.length,
    this.message,
    this.code,
  );
}
